!    Copyright (c) 2009-2010 by High Performance Computing Virtual Laboratory
!
!    This is a race condition example between data updating
!    and referencing in OpenMP and a solution.
!
!    Here we have total number THE_TOTAL_JOBS of independent
!    jobs to complete. They will be grouped into groups with
!    TOTAL_JOBS_PER_GROUP number of jobs in each group. The
!    computation will be done group by group, and inside each
!    group, the "all-slave" model is used. For every group,
!    the initial data Initial_Data_Of_A_Job_Group array is
!    updated just before the group is being performed. Then
!    a race condition occurs when some elements of the
!    Initial_Data_Of_A_Job_Group array are still needed by one
!    thread for his current job, while another thread finds
!    no more jobs left un-assigned in the current group, and
!    wants to update the Initial_Data_Of_A_Job_Group array
!    for the next job group.
!
!    For solutions, please go to norace version.
!
!    www.hpcvl.org
!    Copyright (c) 2009-2010 by High Performance Computing Virtual Laboratory


module all
  use omp_lib
  integer, parameter :: THE_TOTAL_JOBS=6007
  integer, parameter :: TOTAL_JOBS_PER_GROUP=15
  integer            :: Total_Num_of_Groups=(THE_TOTAL_JOBS-1)/TOTAL_JOBS_PER_GROUP+1 
  integer            :: Job_Group=0 
  integer            :: All_Results(THE_TOTAL_JOBS*2) 
  integer            :: Initial_Data_For_Jobs(THE_TOTAL_JOBS) 
  integer            :: Initial_Data_Of_A_Job_Group(TOTAL_JOBS_PER_GROUP) 
  integer            :: End_Of_Group_Jobs=-100 
  integer            :: Job_Assigned=1000, Job_From=100, Job_To=-20 
end module all


program race
  use all
  implicit none
  integer  :: i
  do i=1,THE_TOTAL_JOBS
     Initial_Data_For_Jobs(i)=i;
  end do
!$omp parallel default(shared)
  call Working() 
!$omp end parallel 
  call Check_Results() 
  stop
end program race


subroutine Do_The_Job(my_job)
  use all
  implicit none
  integer  :: my_job, integer
  integer=Initial_Data_Of_A_Job_Group(my_job- &
          ((my_job-1)/TOTAL_JOBS_PER_GROUP)*TOTAL_JOBS_PER_GROUP) 
  All_Results(2*my_job-1)=integer 
  All_Results(2*my_job)=integer*2 
  return
end subroutine Do_The_Job


subroutine Prepare_Job_Group_Data()
  use all
  implicit none
  integer  :: i
  Job_From=(Job_Group-1)*TOTAL_JOBS_PER_GROUP+1
  Job_To=Job_Group*TOTAL_JOBS_PER_GROUP 
  if(Job_To.gt.THE_TOTAL_JOBS) Job_To=THE_TOTAL_JOBS 
  i=Job_To-Job_From+1
  Initial_Data_Of_A_Job_Group(1:i)=Initial_Data_For_Jobs(Job_From:Job_To)
  return
end subroutine Prepare_Job_Group_Data


subroutine Get_Job(the_job)
  use all
  implicit none
  integer  :: the_job
  if(Job_Assigned.lt.Job_To) then
     Job_Assigned=Job_Assigned+1
     the_job=Job_Assigned 
  else 
     the_job=End_Of_Group_Jobs 
  end if
  return
end subroutine Get_Job


subroutine Working()
  use all
  implicit none
  integer  :: my_job
  repeating_do: do
!$omp critical (the_only_one_critical)
      call Get_Job(my_job) 
      if(my_job.eq.End_Of_Group_Jobs) then
         Job_Group=Job_Group+1
         if(Job_Group.le.Total_Num_of_Groups) then
            call Prepare_Job_Group_Data() 
            Job_Assigned=Job_From-1 
            call Get_Job(my_job) 
         end if
      end if
!$omp end critical (the_only_one_critical)
      if(my_job.ne.End_Of_Group_Jobs) then
         call Do_The_Job(my_job) 
      else 
         exit repeating_do
      end if
  end do repeating_do
  return
end subroutine Working  


subroutine Check_Results()
  use all
  implicit none
  integer  :: i,j
  open(9,file="r.txt") 
  do i=1,THE_TOTAL_JOBS
     j=2*i
     write(9,*)i,Initial_Data_For_Jobs(i),All_Results(j-1),All_Results(j),j
     if(i.ne.Initial_Data_For_Jobs(i)) then
        write(9,*)"Initial error: ",i,Initial_Data_For_Jobs(i)
        write(*,*)"Initial error: ",i,Initial_Data_For_Jobs(i)
     end if
     if(i.ne.All_Results(j-1)) then
        write(9,*)"Race ???: ",i,All_Results(j-1)
        write(*,*)"Race ???: ",i,All_Results(j-1)
     end if
     if(All_Results(j).ne.All_Results(j-1)*2) then
        write(9,*)"Computing error: ",All_Results(j-1:j)
        write(*,*)"Computing error: ",All_Results(j-1:j)
     end if
     if(j.ne.All_Results(j)) then
        write(9,*)"Wrong result: ",j,All_Results(j)
        write(*,*)"Wrong result: ",j,All_Results(j)
     end if
  end do
  write(9,*)"Info: ",THE_TOTAL_JOBS,TOTAL_JOBS_PER_GROUP
  write(*,*)"Info: ",THE_TOTAL_JOBS,TOTAL_JOBS_PER_GROUP
  close(9)
  return
end subroutine Check_Results

